
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "amr_plus_fx.h"



#include "typedef.h"
#include "basic_op.h"
#include "oper_32b.h"
#include "count.h"

#define L_FRAME_MAX   (L_FRAME48k)




Word16 Join_over_12k8(/* (o)   : number of sample oversampled       */
  Word16 sig12k8_lf[],/* (i)   : LF signal (fs=12k8)        */
  Word16 sig12k8_hf[],/* (i)   : HF signal (fs=12k8)        */
  Word16 lg,          /* (i)   : length of LF and HF        */
  Word16 sig_fs[],    /* (o)   : oversampled signal         */
  Word16 lg_output,   /* (i)   : L_FRAME44k if 44kHz        */
  Word16 fac_fs,      /* (i)   : >= 32 (32 = base fs)       */
  Word16 mem[],       /* (i/o) : mem[L_MEM_JOIN_OVER]       */
  Word16 *frac_mem    /* (i/o) : interpol fraction memory   */
)
{
  Word16 i, j, ncoef, L_frame, L_frame_int;
  Word16 signal[(2*L_FILT_JOIN)+L_FRAME_PLUS];
  Word16 signal2[(2*L_FILT_OVER)+(2*L_FRAME_PLUS)], *sig;
  Word16 gain, *x1, *x2;
  Word32 s, Ltmp;
  Word16 fac_up, fac_down, tmp16;
  const Word16 *filter;

#ifdef FILTER_44kHz
  test();
  if (L_sub(lg_output,L_FRAME44k) == 0)
  {
    fac_up = 3*441;             move16();
    fac_down = shl(fac_fs,3);        
    filter = (Word16*)Filter_LP165;        move16();
  }
#endif
#ifdef FILTER_48kHz
  test();
  if (L_sub(lg_output,(L_FRAME48k)) == 0)
  {
    fac_up = 180*8;        move16();
    fac_down = shl(fac_fs,3);        
    filter = (Word16 *)Filter_LP180;        move16();
  }
#endif

  ncoef = L_FILT_OVER;              move16();      
  gain = 32767;                     move16();         

  /* frames length */

  L_frame = shl(lg,1);     /* 25k6 rate */     

  /* L_frame = ((L_frame_int*fac_down)+frac) / fac_up */
  /* L_frame_int = ((L_frame*fac_up)-(*frac_mem)+(fac_down-1))/fac_down; */

  Ltmp = L_mult(L_frame, fac_up);
  Ltmp = L_mac(Ltmp, sub(fac_down, add(1, *frac_mem)), 1);

  i = norm_s(fac_down);

  L_frame_int = div_s(extract_h(L_shl(Ltmp, i)), shl(fac_down, i));

  tmp16 = add(L_frame_int, 1);
  Ltmp = L_msu(Ltmp, tmp16, fac_down);
  test();
  if (Ltmp >= 0)
  {
    L_frame_int = tmp16;            move16();
  } 

  /* band join (2 x 12.8kHz --> 25.6kHz) */

  /* load buffer (LF-HF) & update memory */
  for (i=0; i<(2*L_FILT_JOIN); i++) 
  {
    signal[i] = mem[i+(2*L_FILT_OVER)];   move16();
  }
  for (i=0; i<lg; i++) 
  {
    /*signal[i+(2*L_FILT_JOIN)] = sub(sig12k8_lf[i], sig12k8_hf[i]);    move16();*/
    Ltmp = L_mult(sig12k8_lf[i], 32767);
    Ltmp = L_msu(Ltmp, sig12k8_hf[i], 16384);     /* Hy synth is always Q_syn +1*/
    signal[i+(2*L_FILT_JOIN)]  = round(Ltmp);   move16();
  }
  for (i=0; i<(2*L_FILT_JOIN); i++) 
  {
    mem[i+(2*L_FILT_OVER)] = signal[i+lg];    move16();
  }

  /* odd samples = LF-HF interpolated */
  sig = signal2 + (2*L_FILT_OVER);move16();    
  for(i=0; i<lg; i++) 
  {
    x1 = &signal[i+L_FILT_JOIN];move16();    
    x2 = &signal[i+1+L_FILT_JOIN];move16();    
    s = 0;    move32();
    for(j=0; j<L_FILT_JOIN; j++, x1--, x2++) 
    {
      /*s += (*x1 + *x2) * inter2_coef[j];*/
      s = L_mac(s, *x1, Inter2_coef[j]);
      s = L_mac(s, *x2, Inter2_coef[j]);
    }
    sig[(i*2)+1] = round(L_shl(s,1));
  }

  /* synchronise LF+HF with interpolated LF-HF */
  for (i=0; i<(2*L_FILT_JOIN); i++) 
  {
    signal[i] = mem[i+(2*L_FILT_OVER)+(2*L_FILT_JOIN)];   move16();
  }
  for (i=0; i<lg; i++) 
  {
    /*signal[i+(2*L_FILT_JOIN)] = add(sig12k8_lf[i], sig12k8_hf[i]);  move16();*/
    Ltmp = L_mult(sig12k8_lf[i], 32767);
    Ltmp = L_mac(Ltmp, sig12k8_hf[i], 16384);            /* Hy synth is always Q_syn +1*/
    signal[i+(2*L_FILT_JOIN)] = round(Ltmp); move16();

  }
  for (i=0; i<(2*L_FILT_JOIN); i++) 
  {
    mem[i+(2*L_FILT_OVER)+(2*L_FILT_JOIN)] = signal[i+lg];    move16();
  }

  /* even samples = LF+HF */
  for (i=0; i<lg; i++)
  {
    sig[i*2] = signal[i+L_FILT_JOIN];   move16();
  }

  /* load buffer & update memory (for oversampling) */

  for (i=0; i<(2*L_FILT_OVER); i++) 
  {
    signal2[i] = mem[i];            move16();
  }
  for (i=0; i<(2*L_FILT_OVER); i++) 
  {
    mem[i] = signal2[i+L_frame];    move16();
  }

  /* oversample from 25.6kHz to fs */
  Interpol_mem(signal2+ncoef, sig_fs, L_frame_int, filter, ncoef, 
  fac_up, fac_down, gain, frac_mem);
  return(L_frame_int);
}




void Interpol_mem(
  Word16 *signal,         /* (i)  : Signal    */
  Word16 *signal_int,     /* (o)  : interpolated signal           */
  Word16 L_frame_int,     /* (i)  : Length of interpolated signal */
  const Word16 *filter,         /* (i)  : Filter                    Q14 */ 
  Word16 nb_coef,         /* (i)  : Number of coeff               */
  Word16 fac_up,          /* (i)  : Upsampling factor             */
  Word16 fac_down,        /* (i)  : Denumerator on numerator      */
  Word16 gain,            /* (i)  : Gain to apply                 */  
  Word16 *mem_frac        /* (i/o): Fraction memory              */
)
{
  Word16 i, j, frac, frac_step, pos, pos_step;
  Word16 pos_step_plus_one, fac_up_minus_frac_step;
  Word16 frac_d, fac_up_d;
  Word16 *x1, *x2;
  const Word16 *c1, *c2;
  Word32 s;

  /* pos_step = fac_down / fac_up; */
  pos_step = 0;        move16();
  i = fac_down;        move16();
  test();
  while (sub(i, fac_up) >= 0)
  {
    test();
    i = sub(i, fac_up);
    pos_step = add(pos_step, 1);
  }

  pos_step_plus_one = add(pos_step, 1);

  frac_step = sub(fac_down, extract_l(L_shr(L_mult(pos_step,fac_up),1)));
  fac_up_minus_frac_step = sub(fac_up,frac_step);

  fac_up_d = shr(fac_up, 3);

  pos = 0;             move16();
  frac = *mem_frac;    move16();
  for(i=0; i<L_frame_int; i++) 
  {
    x1 = signal+pos;    move16();
    x2 = x1+1;          move16();
    frac_d = shr(frac, 3);
    c1 = &filter[frac_d];           move16();
    c2 = &filter[fac_up_d-frac_d];  move16();
    s = 0;    move32();

    for(j=0; j<nb_coef; j++, c1+=fac_up_d, c2+=fac_up_d) 
    {
      /*s += (*x1--) * (*c1) + (*x2++) * (*c2);*/
      s = L_mac(s, *x1-- , *c1);
      s = L_mac(s, *x2++ , *c2);
    }
    /* in decim and imterpol split, gain is always <= 1*/
    signal_int[i] = shl(mult_r(round(s),gain),1);    move16(); /* Apply gain before left shift to prevent overflow when used in decim_split*/

    test();  
    if (sub(frac,fac_up_minus_frac_step) >= 0) 
    {
      pos = add(pos, pos_step_plus_one);
      frac = sub(frac, fac_up_minus_frac_step);
    }
    else 
    {
      pos = add(pos, pos_step);
      frac = add(frac, frac_step);
    }

  }

  *mem_frac = frac;   move16();
  return;
}


