
#include "typedef.h"
#include "count.h"
#include "basic_op.h"
#include "math_op.h"
#include "oper_32b.h"

#include "amr_plus_fx.h"


#define L_EXTRA    96


static Word16 Short_pitch_tracker(Word16 syn[],Word16 T);                 


void Init_bass_postfilter(Decoder_State_Plus_fx *st)
{
  Set_zero(st->wold_synth_pf, PIT_MAX_MAX+(2*L_SUBFR));
  Set_zero(st->wold_noise_pf, 2*L_FILT);

  st->wold_gain_pf[0] = 0;      move16();
  st->wold_T_pf[0] = 64;        move16();
  st->wold_gain_pf[1] = 0;      move16();
  st->wold_T_pf[1] = 64;        move16();
  
  return;
}


void Bass_postfilter(
  Word16 *synth_in,           /* (i) : 12.8kHz synthesis to postfilter                */
  Word16 *T_sf,               /* (i) : Pitch period for all subframe (T_sf[16])       */
  Word16 *gainT_sf,           /* (i) : Pitch gain for all subframe (gainT_sf[16]) Q14 */ 
  Word16 *synth_out,          /* (o) : filtered synthesis (with delay=L_SUBFR+L_FILT) */
                              /*       delay = (2*L_SUBFR)+L_FILT                     */
  Word16 pit_adj,             
  Decoder_State_Plus_fx *st   /* i/o): decoder memory state                           */
)
{
  Word16 i, j, i_subfr, T, lg, gain, e_tmp, e_ener, m_tmp, m_ener, tmp16, tmp2;
  Word32 Ltmp, Lener;
  Word16 syn_buf[PIT_MAX_MAX+(2*L_SUBFR)+L_FRAME_PLUS], *syn;
  
  Word16 noise_buf[(2*L_FILT)+L_SUBFR], *noise, max, scale, length;

  syn = syn_buf + PIT_MAX_MAX;							move16();
  noise = noise_buf + L_FILT;                           move16();

  T = st->wold_T_pf[0];        move16();
  gain = st->wold_gain_pf[0];  move16();

  test();
  if (pit_adj == 0) 
  {
    Copy(st->wold_synth_pf, syn_buf, PIT_MAX_MAX+L_SUBFR);				
    Copy(synth_in, syn_buf+PIT_MAX_MAX+L_SUBFR, L_FRAME_PLUS);				
    Copy(syn_buf+L_FRAME_PLUS, st->wold_synth_pf, PIT_MAX_MAX+L_SUBFR);		

    Copy(&syn[-L_FILT], synth_out, L_FRAME_PLUS);					
    length = PIT_MAX_MAX + L_SUBFR + L_FRAME_PLUS;    move16();
  }
  else
  {
    Copy(st->wold_synth_pf, syn_buf, PIT_MAX_MAX+(2*L_SUBFR));			
    Copy(synth_in, syn_buf+PIT_MAX_MAX+(2*L_SUBFR), L_FRAME_PLUS);			
    Copy(syn_buf+L_FRAME_PLUS, st->wold_synth_pf, PIT_MAX_MAX+(2*L_SUBFR));		

    Copy(&syn[-L_FILT], synth_out, L_FRAME_PLUS);					
    length = PIT_MAX_MAX + (2*L_SUBFR) + L_FRAME_PLUS;    move16();
  }
  max = 0; move16();
  for(i = 0;i < length; i++)
  {
    test();
    if(sub(abs_s(syn_buf[i]),max)>0)
    {
      max = abs_s(syn_buf[i]);
    }
  }
  scale = sub(norm_s(max),3);  /* 12 bit scaling */
  Scale_sig(syn_buf, length, scale);

  for (i_subfr=0; i_subfr<L_FRAME_PLUS; i_subfr+=L_SUBFR)
  {
    test();
    if (pit_adj != 0) 
    {
      i = sub(shr(i_subfr,6), 2);
      test();
      if (i < 0)
      {
        T = st->wold_T_pf[i+2];       move16();
        gain = st->wold_gain_pf[i+2]; move16();
      }
      else 
      {
        T = T_sf[i];            move16();
        gain = gainT_sf[i];     move16();
      }
    }
   
    test();
    if (sub(gain,16384) > 0) 
    {
      gain = 16384; move16();
    }
    test();
    if (gain < 0) 
    {
      gain = 0;      move16();
    }
    T = Short_pitch_tracker(&syn[i_subfr], T);
    test();
    if (pit_adj == 0)
    {
      lg = sub(L_SUBFR+L_FRAME_PLUS, add(T,i_subfr)); 
    }
    else
    {
      lg = sub(2*L_SUBFR+L_FRAME_PLUS, add(T,i_subfr)); 
    }

    test();
    if (lg < 0)
    {
      lg = 0;  move16();
    }
    test();
    if (sub(lg, L_SUBFR)> 0)
    {
      lg = L_SUBFR; move16();
    }

    /* limit gain to avoid problem on burst */
    test();
    if (lg > 0)
    {
      Ltmp = 0;   move32();
      for (i=0; i<lg; i++) 
      {
        Ltmp = L_mac(Ltmp, syn[i+i_subfr],  syn[i+i_subfr]);
      }
      Lener = 0;     move32();

      for (i=0; i<lg; i++) 
      {
        Lener  = L_mac(Lener, syn[i+i_subfr+T] ,syn[i+i_subfr+T]);

      }
      test();  /* test to avoid division by 0 */
      if (Ltmp ==0)
        Ltmp  = 1;

      /*tmp = (float)sqrt(tmp / ener);     */
      e_tmp   = norm_l(Ltmp);
      e_ener  = norm_l(Lener);
      m_tmp   = extract_h(L_shl(Ltmp,e_tmp));
      m_ener  = extract_h(L_shl(Lener,e_ener));
      
      test();
      if (sub(m_ener, m_tmp)>0)
      {
        m_ener = shr (m_ener, 1);
        e_ener = sub(e_ener,1);
      }
      
      m_ener = div_s(m_ener, m_tmp);
      e_ener = sub(e_ener, e_tmp);
      /* Q15 -> Q31*/
      Ltmp = L_deposit_h(m_ener);
      tmp16 = add(e_ener,16);
      Isqrt_n(&Ltmp, &tmp16);
      /* Output in Q14  (2 bits reserve to gain)*/  
      /* if tmp > gain we don't use */
      /* it and gain <= 1 in Q14 ->(16384)*/
      tmp16 = extract_h(L_shl(Ltmp, add(add(tmp16,e_ener),7)));   
      
      test();
      if (sub(tmp16,gain) < 0)
      {
        gain = tmp16; move16();
      }
    }
    /* calculate noise based on voiced pitch */
    tmp16 = gain; move16(); /*don't need to right shift in fixed-point because gain in Q14*/
    tmp2 = shr(tmp16,1);
    for (i=0; i<lg; i++)
    {
      /*noise_buf[i+(2*L_FILT)] =
        tmp * (syn[i+i_subfr] - 0.5f*syn[i+i_subfr-T] - 0.5f*syn[i+i_subfr+T]);	 */

      Ltmp = L_mult(syn[i+i_subfr], tmp16);
      Ltmp = L_msu(Ltmp, syn[i+i_subfr-T], tmp2);
      Ltmp = L_msu(Ltmp, syn[i+i_subfr+T], tmp2);
      noise_buf[i+(2*L_FILT)] = round(Ltmp);   move16();
      
    }
    for (i=lg; i<L_SUBFR; i++) 
    {
      /*noise_buf[i+(2*L_FILT)] = tmp * (syn[i+i_subfr] - syn[i+i_subfr-T]);	 */
      noise_buf[i+(2*L_FILT)] = round(L_mult(tmp16, sub(syn[i+i_subfr], syn[i+i_subfr-T]))); move16();
    }

    Copy(st->wold_noise_pf, noise_buf, 2*L_FILT);
    Scale_sig(noise_buf, 2*L_FILT, sub(scale, st->Old_bpf_scale));
    
    Copy(noise_buf+L_SUBFR, st->wold_noise_pf, 2*L_FILT);


    /* substract from voiced speech low-pass filtered noise */
    for (i=0; i<L_SUBFR; i++)
    {
      Ltmp = L_mult(Filt_lp[0] ,noise[i]);

      for(j=1; j<=L_FILT; j++)
      {
        Ltmp  = L_mac(Ltmp, Filt_lp[j],  add(noise[i-j] ,noise[i+j]));
      }

      synth_out[i+i_subfr] = sub(synth_out[i+i_subfr], round(L_shr(Ltmp,scale)));	move16(); 
    }

    test();
    if(pit_adj == 0) 
    {
      i = shr(i_subfr,6); 
      T = T_sf[i];          move16();
      gain = gainT_sf[i];   move16();
    }
    /* update last scaling of noise buffer */
    st->Old_bpf_scale = scale;    move16();

  }
  test();
  if (pit_adj == 0)
  {
    st->wold_T_pf[0] = T;        move16();
    st->wold_gain_pf[0] = gain;  move16();  
  }
  else 
  {
    st->wold_T_pf[0]    = T_sf[NB_SUBFR-2];       move16();
    st->wold_gain_pf[0] = gainT_sf[NB_SUBFR-2];   move16();
    st->wold_T_pf[1]    = T_sf[NB_SUBFR-1];       move16(); 
    st->wold_gain_pf[1] = gainT_sf[NB_SUBFR-1];   move16();
  }
  return;
}



static Word16 Short_pitch_tracker(
  Word16 syn[],   /* (i):  synthesis [-PIT_MAX..L_SUBFR] */
  Word16 T        /* (i):  pitch period (>= PIT_MIN)     */
)          
{
  Word16 i,j, T2, ex1, ex2, fr1, fr2;
  Word32 Ltmp, Lcorr, Lener;
  Word32 Ltmp0, Lcorr0, Lener0;
  Word16 *v1, *v2;
  /*----------------------------------------------------------------*
   * Test pitch/2 to avoid continuous pitch doubling                *
   * (short pitch is limited to PIT_MIN (34 = 376Hz) by the encoder *
   *----------------------------------------------------------------*/

  T2 = shr(T, 1);

  v1 = &syn[-L_EXTRA];  move16();
  v2 = &syn[-T2-L_EXTRA];move16();
      

  Lener = 1;   move32();
  Lcorr = 1;   move32();
  Ltmp  = 1;   move32();
  for(j = 0;j < L_SUBFR+L_EXTRA; j+= (L_SUBFR+L_EXTRA)/4)
  {
    Lener0 = 0; move32();
    for (i=0; i<(L_SUBFR+L_EXTRA)/4; i++)
    {
      Lener0  = L_mac(Lener0, v1[i+j],v1[i+j]);
    }
    Lcorr0 = 0; move32();
    for (i=0; i<(L_SUBFR+L_EXTRA)/4; i++) 
    {
      Lcorr0 = L_mac(Lcorr0, v1[i+j], v2[i+j]);
    }
    Ltmp0 = 0; move32();
    for (i=0; i<(L_SUBFR+L_EXTRA)/4; i++)
    {
      Ltmp0 = L_mac(Ltmp0 ,v2[i+j] ,v2[i+j]);
    }
    Lener = L_add(Lener, L_shr(Lener0,1));
    Lcorr= L_add(Lcorr, L_shr(Lcorr0,1));
    Ltmp = L_add(Ltmp, L_shr(Ltmp0,1));

  }
  L_Extract(Lener, &ex1, &fr1);
  L_Extract(Ltmp, &ex2, &fr2);
  Ltmp = Mpy_32(ex1, fr1, ex2, fr2);
  
  Ltmp = L_shr(Ltmp,1);
  Ltmp = Isqrt(Ltmp);
  Lcorr = L_shr(Lcorr,1);
  L_Extract(Ltmp, &ex2, &fr2);
  L_Extract(Lcorr, &ex1, &fr1);
  
  /* cn = normalized correlation of pitch/2 */
  
  Ltmp = Mpy_32(ex1, fr1, ex2, fr2);

  test();
  if (L_sub(Ltmp, 31130)> 0) 
  {
    T = T2;     move16();
  }
  return(T);
}


