#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "typedef.h"
#include "basic_op.h"
#include "count.h"

#include "amr_plus_fx.h"
#include "dec_if_fx.h"
#include "enc_if_fx.h"
#include "wbplus3gplib.h"


extern const UWord8 block_size[];


static void copyright(void) 
{
  fprintf(stderr, "\n");
}


static void usage(char *argv)
{
   fprintf(stderr,
      "Usage: %s [-ff <3gp/raw> -fs <fs_Hz>][-mono] [-limiter] -if <infile.wb+> -of <outfile.wav> -fer <error.txt>\n",
      argv);
   fprintf(stderr, "\n");
   fprintf(stderr, "-ff     file format used (default is 3gp)\n");
   fprintf(stderr, "-fs     frequency sampling\n");
   fprintf(stderr, "-mono   mono output \n");
   fprintf(stderr, "-limiter    avoids output clipping (recommended)\n");
   fprintf(stderr, "-if     input AMRWB+ bitstream file\n");
   fprintf(stderr, "-of     output audio WAV file\n");
   fprintf(stderr, "-fer    frame erasures file (0/1)");
   fprintf(stderr, "\n");
}




static Word32 get_bitrate(DecoderConfig * conf)
{
  Word32 Ltmp;
  Word16 tmp16;
  test();
  if(conf->fscale!=0) 
  {
    tmp16 = mult_r(get_nb_bits(conf->extension, conf->mode, conf->st_mode), 26214); /*Q0*Q21 -> Q6*/
    Ltmp = L_mult(conf->fscale, mult_r(tmp16,21845));   /* Q6*Q6 ->Q13 */
    return Ltmp;
  }
  else 
  {
    return L_shr(L_mult(get_nb_bits(conf->extension, conf->mode, conf->st_mode),26214),9); /*Q22 - 9 ->Q13*/
  }
  
}


static void parsecmdline(
  Word32 argc,
  Word8  *argv[],
  Word8  **input_filename,
  Word8  **output_filename, 
  Word8  **fer_filename, 
  DecoderConfig * conf
  )
{
   conf->fs = 0;                move16();
   conf->mono_dec_stereo = 0;   move16();
   conf->limiter_on = 0;        move16();
   conf->fer_sim = 0;           move16();
   conf->FileFormat = F3GP;     move16();  

   test();
   if (sub((Word16)argc,5) < 0)
   {
      usage(*argv);
      exit(EXIT_FAILURE);
   }
   argc--;
   argv++;
   test();
   while (argc > 0)
   {
      test();test();test();test();test();test();
      if (!strcmp(*argv, "-fs"))
      {
         argv++;
         argc--;
         conf->fs = atoi(*argv);    move16();
      }
      else if (!strcmp(*argv, "-ff"))
      {
         argv++;
         argc--;
         if (!strcmp(*argv, "3gp"))
         {
            conf->FileFormat = F3GP;      move16();
         }
         else
         {
            conf->FileFormat = FRAW;      move16();
         }

      }
      else if (!strcmp(*argv, "-mono"))
      {
         conf->mono_dec_stereo = 1;    move16();
      }
      else if (!strcmp(*argv, "-limiter"))
      {
         conf->limiter_on = 1;    move16();
      }
      else if (!strcmp(*argv, "-if"))
      {
         argv++;
         argc--;
         *input_filename = *argv;    move16();
      }
      else if (!strcmp(*argv, "-of"))
      {
         argv++;
         argc--;
         *output_filename = *argv;    move16();
      }
      else if (!strcmp(*argv, "-fer"))
      {
         argv++;
         argc--;
         *fer_filename = *argv;       move16();
         conf->fer_sim=1;             move16();
      }
      else
      {
         fprintf(stderr, "Unknown option %s\n", *argv);
         exit(EXIT_FAILURE);
      }
      argv++;
      argc--;

   }
}

static void select_fs(DecoderConfig* conf) 
{

  /* default sampling rate if undefined */
  test();
  if (conf->fs == 0)
  { test();test();
    if (conf->extension == 0)
    {
      conf->fs=16000;       move32(); /*16khz for AMRWB*/
    }
    else if (conf->fscale == 0)
    {
      conf->fs=24000;      /*24khz*/      move32();
    }
    else
    {
      conf->fs = 48000;   move32();                        /*48kHz*/
    }
  } 

  /* user specified sampling rate */
  else 
  {
    test();
    if(conf->extension == 0) 
    { test();
      if(L_sub(conf->fs,16000) != 0) 
      {
        fprintf( stderr, "Sampling rate not supported" );
        exit(EXIT_FAILURE);
      }
    } 
    else
    { test();
      if (conf->fscale == 0) 
      { test();test();test();
        if ( (L_sub(conf->fs,16000) != 0) && (L_sub(conf->fs,24000) != 0) && (L_sub(conf->fs,8000) != 0)) 
        {
          fprintf( stderr, "Sampling rate not supported" );
          exit(EXIT_FAILURE);
        }
      }
      else
      { test();
        if ((L_sub(conf->fs,44100)!=0) && (L_sub(conf->fs,48000) != 0) ) 
        {
          fprintf( stderr, "Sampling rate not supported" );
          exit(EXIT_FAILURE);
        }
      }
    }
  }
}
static void set_frame_length(
  DecoderConfig* conf,
  Word16 *L_frame
)
{
  
  test();test();test();test();test();
  move16();
  switch (conf->fs)
  {
      
    case 8000:
      *L_frame = L_FRAME8k;
      break;
      
    case 16000:
      *L_frame = L_FRAME16kPLUS;
      
      break;
      
    case 24000:
      *L_frame = L_FRAME24k;
      break;
    case 32000:
      *L_frame = L_FRAME32k;
      break;
#ifdef FILTER_44kHz
    case 44100:
      *L_frame = L_FRAME44k;     move16();
      break;
#endif
#ifdef FILTER_48kHz
    case 48000:
      *L_frame = L_FRAME48k;
      
      break;
#endif
      
    default:
      fprintf( stderr, "Sampling rate not supported" );
      exit(EXIT_FAILURE);
  }
}

static void Interleave(
  Word16 *right,
  Word16 *left,

  Word16 *out,
  Word16 length
)
{
  Word16 i;

  for (i=sub(length,1); i>=0; i--) 
  {
    out[(i*2)+1] = right[i];    move16();
    out[i*2] = left[i];         move16();
  }

}



static void Simple_frame_limiter(    
  Word16 x[],    /* i/o: synthesis                        Qx */
  Word16 Qx,     /* i:   scale factor of x                  */
  Word16 mem[],  /* i/o: mem[2] should be init to [0,0]     */
  Word16 n       /* i:   vector size                        */
)
{
 
  Word16 i, fac, prev, tmp, frame_fac, max;
  Word32 L_tmp;

  /*------------------------------------------------------------------------------------------*
  * calculate frame AGC factor to avoid saturation 
  *------------------------------------------------------------------------------------------*/

  max = 0;                                    move16();
  for (i=0; i<n; i++)
  {
    tmp = abs_s(x[i]);

    test();
    if (sub(tmp, max) > 0)
    {
      max = tmp;                          move16();
    }
  }

  tmp = shl(30000, Qx);      /* saturation can occurs here */

  frame_fac = 0;                              move16();
  test();
  if (sub(max, tmp) > 0)
  {
    frame_fac = sub(16384, div_s(shr(tmp, 1), max));   /* frame fac in Q15 */
  }

  fac = mem[0];                               move16();   /* fac in Q15 */
  prev = mem[1];                              move16();   /* prev in Q-1 */

  /*------------------------------------------------------------------------------------------*
   * AGC
   *------------------------------------------------------------------------------------------*/

  for (i=0; i<n; i++)
  {
    /* update AGC factor (slowly) */
    fac = round(L_mac(L_mult(32440, fac), 328, frame_fac));

    /* convert float to integer with AGC */
    L_tmp = L_deposit_h(x[i]);
    L_tmp = L_msu(L_tmp, fac, x[i]);
    L_tmp = L_msu(L_tmp, fac, prev);
    L_tmp = L_shr(L_tmp, Qx);    /* saturation can occurs here */

    prev = x[i];                            move16();
    x[i] = round(L_tmp);                    move16();


  }

  mem[0] = fac;                               move16();
  mem[1] = prev;                              move16();

  return;
}


long ext_frame = 0;

void main(
  int argc,
  char *argv[]
)
{
  FILE *f_serial;       /* File of serial bits for transmission  */
  FILE *f_syn;          /* File of synthesis speech      */
  FILE *f_fer;          /* File of frame ereasures*/
  
  Word16 speech16[L_FRAME16k];
  Decoder_State_Plus_fx *st_d;
  void *stAmrwbDec;
  UWord16 serial[NBITS_MAX];         /* serial parameters.   */
  Word16 bitsPerSample = 16; 
  Word16 bfi[4]= {0,0,0,0};
  Word16 i, j, nb_bits, L_frame;
  Word32 frame,  dataSize;
  
  Word16 channel_right[4*L_FRAME48k];
  Word16 channel_left[2*L_FRAME48k];
  Word16 nb_samp;
  Word16 mode, extension;
  Word16 st_mode;
  Word16 num_channels;

  Word16 mem_down_right[2*L_FILT_DECIM_FS], mem_down_left[2*L_FILT_DECIM_FS];
  Word16 frac_down_right, frac_down_left;
  Word16 fac_up, fac_down, nb_samp_fs, last_scaling;
  Word32 fs_output;
  
  Word8 *input_filename="";
  Word8 *output_filename="";
  Word8 *fer_filename="";
  Word16 fst_switched;

  /* frame limiter*/    
  Word16 Frame_Q;  
  Word16 mem_fr_lim_right[2];
  Word16 mem_fr_lim_left[2];
  Word16 old_st_mode;
  Word16 fst;

  DecoderConfig conf;
  Word16 left_limiter_gain;
  Word16 right_limiter_gain;

  Word16 nb_byte;
  
  Word16 tfi;

  Word32 old_bitrate, tmpbfi;  
  
  left_limiter_gain = 32767;
  right_limiter_gain= 32767;
  fst = 0;
  mode = 0;
  nb_bits = 0;
  Frame_Q = 0;
  
  /* Display Copyright Banner*/
  
  copyright();
  
  /* Parse Command Line */

  parsecmdline(argc, argv, &input_filename, &output_filename,&fer_filename, &conf);
  
  
  /* Open input bitstream file*/
  
  if(conf.FileFormat == F3GP)
  {
      if(Open3GP(&tfi, bfi, input_filename, 0, &conf))
      {
          fprintf(stderr, "Error opening the input bitstream file %s.\n",
                input_filename);
          exit(EXIT_FAILURE);
      }
  }
  else
  {
      if ((f_serial = fopen(input_filename, "rb")) == NULL)
      {
          fprintf(stderr, "Error opening the input bitstream file %s.\n",
                input_filename);
          exit(EXIT_FAILURE);
      }
      ReadHeader(&tfi, bfi, &(conf.extension), &(conf.mode), &(conf.st_mode), &(conf.fscale), f_serial);
      rewind(f_serial);
  }

  /* In case of bad frame; hard initialization */
  test();test();test();test();
  if((L_sub(conf.fs,16000) == 0|| L_sub(conf.fs, 8000) == 0) && sub(conf.mode,14) == 0)
  {   
     /* Force to wb */ 
      conf.fscale = 0;      move16();
      conf.extension = 0;   move16();
  }
  else if ( sub(conf.mode,14) == 0)
  {
     /* Force to wb+ if sampling rate desire is different than 16 or 8kHz */
      conf.mode = 2;      move16();
      conf.fscale = 96;   move16();
      conf.extension = 1; move16();
  }

  /* default sampling rate if undefined */
  test();
  if (conf.fs == 0)
  {
    test();test();
    if (conf.extension == 0)
    {
      conf.fs = 16000;     move16();
    }
    else if (conf.fscale == 0)
    {
      conf.fs = 24000;   move16();
    }
    else
    {
#ifdef FILTER_44kHz
      conf.fs = 44100;   move16();
#endif
#ifdef FILTER_48kHz
      conf.fs = 48000;   move16();
#endif
    }
  }
  fs_output = conf.fs;   move16();

  fac_up = fac_down = 12;     move16(); /* no oversampling by default */

  test();
  if (conf.fscale != 0)
  {
    test();test();test();test();test();test();
    switch (conf.fs) 
    {
      case 8000:
        fac_up = 2;      move16();
        conf.fs = 48000;   move16();
      break;
      case 16000:
        fac_up = 4;   move16();
        conf.fs = 48000;   move16();
      break;
      case 24000:
        fac_up = 6;   move16();
        conf.fs = 48000;   move16();
      break;
      case 32000:
        fac_up = 8;   move16();
        conf.fs = 48000;   move16();
      break;
      case 11025:
        fac_up = 3;   move16();
        conf.fs = 44100;   move16();
      break;
      case 22050:
        fac_up = 6;   move16();
        conf.fs = 44100;   move16();
      break;
    }
    Set_zero(mem_down_right, 2*L_FILT_DECIM_FS);
    Set_zero(mem_down_left, 2*L_FILT_DECIM_FS);
  }
  frac_down_right = 0;     move16();
  frac_down_left = 0;      move16();
  last_scaling = 0;        move16();

  /* select sampling rate */
  select_fs(&conf);
  
  /* Set frame lengths and check for sampling frequency */
  set_frame_length(&conf,&L_frame);
  
  num_channels = 2;     move16();
  test();test();test();test();
  if ( (conf.fscale == 0) && (conf.st_mode < 0)
       || (conf.extension == 0) 
       || (conf.mono_dec_stereo))

  {
    num_channels = 1;     move16();
  }

  /* Frame errasure simualtion */
  test();
  if(conf.fer_sim > 0) 
  {
      test();
      if ((f_fer = fopen(fer_filename,"rt")) == NULL) 
      {
         fprintf(stderr,"Error opening fer file %s .\n",fer_filename);
         exit(0);
      }
  }

  if ((f_syn = Wave_Fopen(output_filename, "wb", &num_channels, &fs_output,
          &bitsPerSample, &dataSize)) == NULL)
  {
    fprintf(stderr, "Error opening output wav file %s.\n", output_filename);
    exit(EXIT_FAILURE);
  }
  /*
   Display Information
  */
  old_bitrate = get_bitrate(&conf);
  fprintf(stderr,"Decoding @ %6.2fkbps",(float)old_bitrate/8192.0);
  


  stAmrwbDec = D_IF_init_fx();
  
  st_d = malloc(sizeof(Decoder_State_Plus_fx));
  test();
  if(st_d == NULL)
  {
    fprintf(stderr, "Can't allocate st_d\n" );
    exit(EXIT_FAILURE);
  }
  Init_decoder_amrwb_plus(st_d, (Word16)num_channels, conf.fscale, 1);
  Init_WMOPS_counter ();

  fprintf(stderr,"\n --- Running ---\n");
  
  
  /*---------------------------------------------------------------------------*
   * Loop for every analysis/transmission frame.                               *
   *   -New L_FRAME_PLUS data are read. (L_FRAME_PLUS = number of speech data per frame) *
   *   -Conversion of the speech data from 16 bit integer to real              *
   *   -Call coder_wb to encode the speech.                                    *
   *   -The compressed serial output stream is written to a file.              *
   *   -The synthesis speech is written to a file                              *
   *--------------------------------------------------------------------------*/
  
  frame = 0;

  /*  n_samples = 0;*/
  old_st_mode = conf.st_mode;
  extension = conf.extension;
  mode = conf.mode;
  st_mode = conf.st_mode;
  fst = conf.fscale;
  tfi = (Word16)(frame%4);  logic16();

  Set_zero(mem_fr_lim_right,2);    
  Set_zero(mem_fr_lim_left,2);
  nb_byte = 0;
  while (1)   
  {

    fst_switched = 0;     move16();
    fprintf(stderr,"Frames processed: %ld    \r", frame/4);
    Reset_WMOPS_counter();
    /* read 4 packets of 20ms with header */
    /* if mode==15 (no data) this mean that the frame is corrupted (20ms lost)
         and the decoder will try to recover signal with others packets */

    /*  Get next frame 4x20ms if AMR-WB+ or 1x20ms if AMR-WB */
    test();
    if (conf.FileFormat == F3GP)
    {
        if(GetNextFrame3GP(&tfi, bfi, &extension,&mode,&st_mode,&fst,(void *)serial,0)) 
        {
           break;  
        }
        conf.mode = mode;
        conf.st_mode = st_mode; 

    }
    else
    {
        if(!ReadRawFile(&tfi, bfi, &conf, &extension, &mode, &st_mode, &fst, f_serial, (void*) serial))
          break;

    }
    /* set the bfi for SIMULATION PURPOSE */
    /* read frame erasures every forth frame for wb+*/
    /* read frame erasures every frame for amr-wb*/
    test();
    if(conf.fer_sim) 
    {  
        if(conf.extension > 0) 
         {
             for(i=0;i<4;i++) 
             {
                fscanf(f_fer,"%d",&tmpbfi);
                bfi[i] = tmpbfi;
             }
         }
         else
         {
             fscanf(f_fer,"%d",&tmpbfi);         
             bfi[tfi] = tmpbfi;

         }
    }
    test();test();
    if ( (sub(fst,conf.fscale ) != 0) && !fst_switched) 
    {
      select_fs(&conf);
      set_frame_length(&conf, &L_frame);
      fst_switched = 1;   move16();
      Init_decoder_amrwb_plus(st_d, num_channels, fst, 0);
    }
        
        
    /* SET CONFIG !!!!!!! */
    test();test();test();test();
    if(((extension==0) && (sub(conf.extension,1)==0))
        || ((sub(extension,1)==0) && (conf.extension==0)))
    {
        
      /* Copy decoder data in case of switch between WB and WB+ modes*/
      test();test();test();test();test();test();test();test();
      if ( (( (mode >= 0) && (sub(mode,9) <= 0) ) || (sub(mode,15) == 0)) && (conf.extension>0) )
      {
        /* Switch from WB+ to WB*/
        Copy_decoder_state(st_d, stAmrwbDec, 1);
      }
      else if ( ((mode >= 0) && (sub(mode,8) <= 0)|| (sub(mode,15) == 0)) && (conf.extension==0) )
      {
        /* Switch from WB to WB+*/
        Copy_decoder_state(st_d, stAmrwbDec, 0);
      }
        
      conf.mode = mode;             move16();
      conf.extension = extension;	  move16();
    }
    
    test();
    if(conf.extension > 0) 
    {
        nb_samp = Decoder_amrwb_plus(conf.mode, (Word16*)serial, bfi, L_frame,
                        num_channels, channel_right, channel_left, st_d,
                        fst, conf.st_mode, conf.mono_dec_stereo, conf.fscale);
      Frame_Q = st_d->Old_Q_syn;    move16();
      frame +=4;
    }
    else 
    {
      nb_samp = L_FRAME16k;     move16();
      D_IF_decode_fx(stAmrwbDec, (UWord8*)serial, speech16,bfi[tfi]? _bad_frame: _good_frame);
      Frame_Q = 0;    move16();

      frame++;
      for (j = 0; j < 320; j++)
      {
        channel_right[ j] = speech16[j];    move16();
      }
    }
      
    old_st_mode = conf.st_mode; move16();

    test();
    if (sub(num_channels,2) == 0)
    {
      /* Smooth the output so to avoid harch clipping*/
      test();
      if(conf.limiter_on>0)
      {
        Simple_frame_limiter(channel_right,Frame_Q, mem_fr_lim_right, nb_samp);
        Simple_frame_limiter(channel_left,Frame_Q, mem_fr_lim_left, nb_samp);
        Frame_Q = 0;    move16();
      }
      Scale_sig(mem_down_right,2*L_FILT_DECIM_FS, sub(Frame_Q,last_scaling));  
      Scale_sig(mem_down_left,2*L_FILT_DECIM_FS, sub(Frame_Q,last_scaling));  

      last_scaling = Frame_Q;   move16();

      nb_samp_fs = Decim_fs(channel_right, nb_samp, channel_right, fac_up, mem_down_right, &frac_down_right);
      nb_samp_fs = Decim_fs(channel_left, nb_samp, channel_left, fac_up, mem_down_left, &frac_down_left);

            /* interleave of left and right samples (stereo file format) */
      Interleave(channel_right,channel_left,channel_right,nb_samp_fs);
      Writ_data(channel_right, shl(nb_samp_fs,1), f_syn, Frame_Q);
    } 
    else
    {
      test();test();
      if(conf.limiter_on>0) 
      {
        Simple_frame_limiter(channel_right,Frame_Q, mem_fr_lim_right, nb_samp);
        Frame_Q = 0;    move16();
        move16();
      }
      Scale_sig(mem_down_right,2*L_FILT_DECIM_FS, sub(Frame_Q, last_scaling));  
      last_scaling = Frame_Q;   move16();

      nb_samp_fs = Decim_fs(channel_right, nb_samp, channel_right, fac_up, mem_down_right, &frac_down_right);
      Writ_data(channel_right, nb_samp_fs, f_syn, Frame_Q);
    }
    
    conf.fscale = fst;   move16();
    test();
    if (fst_switched>0)
    {
      set_frame_length(&conf,&L_frame);

    }
  
    /* Flush file stream*/

    (void)fflush(f_syn);   
    test();test();
    if(conf.mode != 14 && conf.mode != 15) /* prevent wrong bitrate info */
    {
        test();    
        if (L_sub(L_abs(L_sub(old_bitrate, get_bitrate(&conf))),1) > 0) 
        {
            old_bitrate = get_bitrate(&conf);
            test();
            if (old_bitrate != 0)
                fprintf(stderr, "Decoding @ %6.2fkbps\n", (float)old_bitrate/8192.0f);
            else /* DTX frame return a bit rate of 0 kbps */
                fprintf(stderr, "Decoding         DTX\n");
        }
    }
  }  
  free(st_d);
  if (fst == 0) fst = FSCALE_DENOM;
  WMOPS_output(1, (float)fst / FSCALE_DENOM );

  Wave_Fclose(f_syn, bitsPerSample);

  /* Output Complexity counts*/
  exit(EXIT_SUCCESS);
}


